# compat.py
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# Authors:
#     Trae Santiago <trae32566@gmail.com>
'''Compatibility mode handler for Safestore with older Python

Designed to be as modular as possible, so that alternate backup methods can be added in the future. 
This handles both parsing of the config and calling of all the workers.
'''

from datetime import datetime
from glob import glob
from imp import find_module, load_module
from os import getcwd, listdir
from os.path import exists, isdir, ismount, join, sep
from pprint import pformat
from Queue import Queue, Empty
from shutil import rmtree
from socket import gethostname
from subprocess import CalledProcessError, check_output, STDOUT
from sys import exc_info
from threading import Event, Thread
from time import gmtime, strftime

# Local imports (yes, globs!)
from .base import AgnosticBase
from .exceptions import *

class Core(AgnosticBase):
    '''Core threaded (but not truly) handler for Safestore with older Python installations

    This class uses threading instead of asyncio, but we're offering both right now due to limited
    availability of Python 3 in distros. The performance difference for now is negligible, however,
    asyncio does not have any thread switching overhead for highly parallel I/O operations, such as
    large storage arrays that have high throughput.
    '''
    #########################
    #     Backup stuffs     # 
    #########################
    def backup(self, directArgs=None):
        '''Backup handler, since _backup_worker() needs to be managed.

        Arguments:
        directArgs - A list of direct arguments to pass to every rsync instance. This overrides the
        default arguments that we give it, so make sure to include arguments needed, EXCEPT THOSE 
        REQUESTED BY CONFIGURATION PARAMETERS (ie: bwCap). As well, the first (parent) queue item 
        created for asynchronous backups will have an additional flag, so that it ignores all 
        subdirectories. All subsequent rsync calls will not have this.

        A few caveats here:
        o There is no guarantee that this will not cause system instability due to high load
        o Eventually an _optimize() method will be added to calculate the ideal load distribution,
            but for now we're just using a basic top level directory crawling method.
        '''
        self.deadTree.info('Starting backup activity')

        # Make a queue for backup tasks
        self.bkpQueue = Queue()

        # Create an event to let the workers know when to die
        self.bkpComplete = Event()

        # Set a consistent start time
        self.startTime = strftime('%Y-%m-%d_%H-%M', gmtime())

        self.rsyncChanges = ''

        for source, bkpConfig in self.config['backup'].items():
            # Make sure the event is not complete yet
            if self.bkpComplete.is_set(): self.bkpComplete.clear()

            # Plugin stuffs
            # Exception handling here is CRITICAL, we need to make sure these plugins fail in a 
            # predictable way to prevent Safestore from exploding!
            if source[0] == ':' and source[-1] == ':':
                # Module location for installations
                pluginPath = (
                    '{0}/safestore'.format(getcwd())
                    if exists('{0}/safestore'.format(getcwd())) else 
                    '/usr/lib/safestore/safestore'
                )

                try:
                    modInfo = find_module(source[1:-1], [pluginPath])
                    pluginMod = load_module(source[1:-1], *modInfo)
                except ImportError as err:
                    self.deadTree.critical('Loading plugin {0} failed!'.format(source[1:-1]))
                    self.deadTree.debug('Error was:{0}\n'.format(pformat(err)))
                    exit(40)

                for method, objName in bkpConfig.items():
                    try:
                        # Pass everything after the plugin name
                        pluginMod.plugin_wrapper(
                            method, 
                            objName.keys()[-1], 
                            self.config['backup'][source], 
                            self.deadTree
                        )
                        self.deadTree.info('{0} complete!'.format(source[1:-1]))
                    except:
                        raise PluginError(exc_info(), source[1:-1], self.deadTree)
            else:
                # Set async defaults
                if not 'async' in bkpConfig: bkpConfig['async'] = {}
                if not 'maxThreads' in bkpConfig['async']: bkpConfig['async']['maxThreads'] = 1
                if not 'pollInterval' in bkpConfig['async']: bkpConfig['async']['pollInterval'] = 0.01
                
                # Warn if we don't have a trailing '/'
                if source[-1] != sep:
                    self.deadTree.warning(
                        "Trailing separator missing! This might not be what you want!"
                    )

                for count in range(bkpConfig['async']['maxThreads']):
                    bkpThread = Thread(
                        target=self._backup_worker,
                        args=(bkpConfig['async']['pollInterval'],)
                    )
                    bkpThread.daemon = True
                    bkpThread.start()

                flags = ['-abixzAS', '--delete'] if directArgs is None else directArgs

                # Single threaded (any source which is not a file, ie: rsync URI, is not async!)
                if bkpConfig['async']['maxThreads'] == 1 or not isdir(source):
                    self.bkpQueue.put((source, None, flags))
                # Multi-threaded
                else:
                    # The parent portion here is mandatory
                    parentFlags = flags[:]
                    parentFlags.append('--filter=-,p */*')

                    # Make sure we run the parent dir backup FIRST
                    self.bkpQueue.put((source, None, parentFlags))
                    self.bkpQueue.join()
                    if hasattr(self, 'explosion'): raise self.explosion                

                    # Crawl the top level (this is very basic and limits performance severely with few base dirs)
                    # STEALME
                    for item in listdir(source):
                        fullPath = join(source, item)

                        if isdir(fullPath) and not ismount(fullPath):
                            item += sep
                            self.bkpQueue.put((source, item, flags))

                    # Try optimizing here perhaps
                    #self.optimize(source)

            # *default* is serial. Don't get confused!
            if not 'safestore' in self.config or not 'serial' in self.config['safestore'] or self.config['safestore']['serial']:
                self.bkpQueue.join()
                if hasattr(self, 'explosion'): 
                    raise self.explosion                
                else:
                    self.bkpComplete.set()
                    self.deadTree.info('Serial backup complete!')
            else:
                self.deadTree.info('Workers started for this backup!')

        if 'safestore' in self.config and 'serial' in self.config['safestore'] and not self.config['safestore']['serial']:
            self.bkpQueue.join()
            if hasattr(self, 'explosion'):
                raise self.explosion
            else:
                self.bkpComplete.set()
                self.deadTree.info('Backup complete!') 

    def _backup_worker(self, timeout):
        '''The "meat" of the backup process

        This is an internal method called by Core().backup() for each thread instance.
        Note: This uses --exclude, NOT --exclude-from=- because passing stdin is a pain.
        '''
        while not self.bkpComplete.is_set():
            try:
                source, subDir, directArgs = self.bkpQueue.get(True, timeout)
            except Empty:
                continue

            # We must, indeed, all hang together or, most assuredly, we shall all hang separately
            try:
                # Leave the path alone if we don't specify a subdir
                fullDir = source if subDir is None else join(source, subDir)

                destDir = (
                    self.config['backup'][source]['destination'] 
                    if subDir is None 
                    else '{0}{1}{2}'.format(self.config['backup'][source]['destination'], sep, subDir[:-1])
                )

                prefix = (
                    gethostname().split('.')[0] if not 'prefix' in self.config['backup'][source]
                    else self.config['backup'][source]['prefix']
                )
                deltaDir = '{0}_{1}'.format(prefix, self.startTime)

                protectDir = (
                    gethostname().split('.')[0]
                    if not 'prefix' in self.config['backup'][source]
                    else self.config['backup'][source]['prefix']
                )

                self.deadTree.info("Copying '{0}'".format(fullDir))
                rsyncArgs = [
                    'rsync',
                    '--filter=protect {0}*'.format(protectDir),
                    fullDir,
                    destDir
                ]

                # We don't always want deltas
                if not 'deltas' in self.config['backup'][source] or self.config['backup'][source]['deltas']:
                    rsyncArgs.insert(-2, '--backup-dir={0}'.format(deltaDir))
                if 'bwCap' in self.config['backup'][source]:
                    rsyncArgs.insert(
                        -2,
                        '--bwlimit={0}'.format(self.config['backup'][source]['bwCap'])
                    )

                # Password file
                if 'passFile' in self.config['backup'][source]:
                    rsyncArgs.insert(
                        -2,
                        '--password-file={0}'.format(self.config['backup'][source]['passFile'])
                    )

                # Exclusion list
                if 'excludes' in self.config['backup'][source]:
                    for exclude in self.config['backup'][source]['excludes']:
                        if subDir is None:
                            rsyncArgs.insert(-2, '--exclude={0}'.format(exclude))
                        elif exclude.find(subDir) != -1:
                            rsyncArgs.insert(-2, '--exclude={0}'.format(exclude.replace(subDir, '', 1)))

                # Add arguments
                [ rsyncArgs.insert(1, directArg) for directArg in directArgs ]

                self.deadTree.debug('subprocess.Popen() command list:\n{0}'.format(rsyncArgs))

                # We're using close_fds=True to work around subprocess.Popen() not being thread safe:
                # https://stackoverflow.com/questions/21194380/is-subprocess-popen-not-thread-safe#comment31948801_21194380
                rsyncInstance = check_output(rsyncArgs, close_fds=True, stderr=STDOUT)
                self.deadTree.info("Finished copying '{0}'".format(fullDir))
            except CalledProcessError as err:
                if err.returncode in self.retWarnActions:
                    self.deadTree.warning(self.retWarnActions[err.returncode])
                    self.deadTree.debug('Rsync output was:\n'.format(err.output))
                else:
                    self.explosion = RsyncError(
                        err,
                        self.deadTree,
                        killEvent=self.bkpComplete, 
                        childQueue=self.bkpQueue
                    )
            except OSError as err:
                self.deadTree.debug('Error code: {0}'.format(err.errno))
                self.deadTree.debug('Error description: {0}'.format(err.strerror))
                if hasattr(err, 'filename'): self.deadTree.debug('File: {0}'.format(err.filename))
                self.explosion = UnknownChildError(
                    self.bkpComplete, 
                    self.bkpQueue, 
                    exc_info(),
                    self.deadTree 
                )
            except:
                self.explosion = UnknownChildError(
                    self.bkpComplete, 
                    self.bkpQueue, 
                    exc_info(),
                    self.deadTree 
                )

            self.bkpQueue.task_done()

    #########################
    #     Rotate stuffs     # 
    #########################
    def rotate(self):
        '''Rotate wrapper, since _rotate_worker() needs to be queued.

        There is no guarantee that this will not cause system instability due to high load.
        '''
        self.deadTree.info('Starting rotation activity')

        # Make a queue
        self.rotQueue = Queue()

        # Create an event to let the workers know when to die
        self.rotComplete = Event()

        for source, rotConfig in self.config['rotate'].items():
            # Make sure the event is not complete yet
            if self.rotComplete.is_set(): self.rotComplete.clear()

            # Set async defaults
            if not 'async' in rotConfig: rotConfig['async'] = {}
            if not 'maxThreads' in rotConfig['async']: rotConfig['async']['maxThreads'] = 1
            if not 'pollInterval' in rotConfig['async']: rotConfig['async']['pollInterval'] = 0.01

            for count in range(rotConfig['async']['maxThreads']):
                rotThread = Thread(
                    target=self._rotate_worker,
                    args=(rotConfig['async']['pollInterval'],)
                )
                rotThread.daemon = True
                rotThread.start()

            # Single threaded
            if rotConfig['async']['maxThreads'] == 1:
                self.rotQueue.put((source, None))
            # Multi-threaded
            else:
                # Crawl the top level
                # Needs to be replaced with Core.schedule()
                for item in listdir(source):
                    if isdir(join(source, item)): self.rotQueue.put((source, item))

            if not 'serial' in self.config['safestore'] or self.config['safestore']['serial']:
                self.rotQueue.join()
                if hasattr(self, 'explosion'):
                    raise self.explosion
                else:
                    self.rotComplete.set()
                    self.deadTree.info('Serial rotation complete!')
            else:
                self.deadTree.info('Workers started for this rotation!')

        if 'serial' in self.config['safestore'] and not self.config['safestore']['serial']:
            self.rotQueue.join()
            if hasattr(self, 'explosion'):
                raise self.explosion
            else:
                self.rotComplete.set()
                self.deadTree.info('Rotation complete!')

    def _rotate_worker(self, timeout):
        '''The "meat" of the rotation process

        This is an internal method called by Core().rotate() for each thread instance.
        '''
        while not self.rotComplete.is_set():
            try:
                source, subDir = self.rotQueue.get(True, timeout)
            except Empty:
                continue

            # We catch all exceptions, then pass them to the parent to be raised, thus preventing 
            # hung threads
            try:
                fullDir = join(source, subDir)

                for clientDir in listdir(fullDir):
                    self.deadTree.info("Rotating '{0}'".format(join(fullDir, clientDir)))

                    deltaDirs = glob('{0}/{1}*/'.format(join(fullDir, clientDir), subDir))
                    if deltaDirs:
                        deltaDirs.sort()
                        excess = len(deltaDirs) - int(self.config['rotate'][source]['count'])

                        while excess > 0:
                            self.deadTree.info(
                                '{0} excess backup(s), removing: {1}'.format(excess, deltaDirs[0])
                            )
                            rmtree(deltaDirs[0])

                            # Decrement
                            excess -= 1
                            deltaDirs.pop(0)
            except: 
                self.explosion = UnknownChildError(
                    self.rotComplete, 
                    self.rotQueue, 
                    exc_info(),
                    self.deadTree 
                )

            self.rotQueue.task_done()