# core.py
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the 
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program.  
# If not, see <http://www.gnu.org/licenses/>.
# Authors:
#     Trae Santiago <trae32566@gmail.com>
'''Core asyncronous handler for Safestore

Designed to be as modular as possible, so that alternate backup methods can be added in the 
future. This handles both parsing of the config and calling of all the workers.
'''

import asyncio
from datetime import datetime
from glob import glob
from imp import find_module, load_module
from os import getcwd, listdir
from os.path import exists, isdir, ismount, join, sep
from pprint import pformat
from shutil import rmtree
from socket import gethostname
from subprocess import CalledProcessError, check_output, STDOUT
from sys import exc_info
from time import gmtime, strftime

# Local imports (yes, globs!)
from .base import AgnosticBase
from .exceptions import *

class Core(AgnosticBase):
    '''Core asyncronous handler for Safestore with Python 3 installations

    This class uses asyncio instead of threading for better performance, but only for Python 3.4+.
    '''
    def backup(self, directArgs=None):
        '''Backup handler, since _backup_worker() needs to be managed.
        
        Arguments:
        directArgs - A list of direct arguments to pass to every rsync instance. This overrides the
        default arguments that we give it, so make sure to include arguments needed, EXCEPT THOSE 
        REQUESTED BY CONFIGURATION PARAMETERS (ie: bwCap). As well, the first (parent) queue item 
        created for asynchronous backups will have an additional flag, so that it ignores all 
        subdirectories. All subsequent rsync calls will not have this.

        A few caveats here:
        o There is no guarantee that this will not cause system instability due to high load
        o Eventually an _optimize() method will be added to calculate the ideal load distribution,
            but for now we're just using a basic top level directory crawling method.
        '''
        self.deadTree.info('Starting backup activity')

        # Make a queue for backup tasks
        self.bkpQueue = asyncio.Queue()

        # Create an asyncio event loop
        self.bkpLoop = asyncio.get_event_loop()
        
        # Set a consistent start time
        self.startTime = strftime('%Y-%m-%d_%H-%M', gmtime())

        for source, bkpConfig in self.config['backup'].items():
            # Plugin stuffs
            if source[0] == ':' and source[-1] == ':':
                # Module location for installations
                pluginPath = (
                    '{0}/safestore'.format(getcwd())
                    if exists('{0}/safestore'.format(getcwd())) else 
                    '/usr/lib/safestore/safestore'
                )

                try:
                    modInfo = find_module(source[1:-1], [pluginPath])
                    pluginMod = load_module(source[1:-1], *modInfo)
                except ImportError as err:
                    self.deadTree.critical('Loading plugin {0} failed!'.format(source[1:-1]))
                    self.deadTree.debug('Error was:{0}\n'.format(pformat(err)))
                    exit(20)

                for method, objName in bkpConfig.items():
                    # Pass everything after the plugin name
                    pluginMod.plugin_wrapper(
                        method, 
                        objName.keys()[-1], 
                        self.config['backup'][source], 
                        self.deadTree
                    )
                    self.deadTree.info('{0} complete!'.format(source[1:-1]))
            else:
                # Warn if we don't have a trailing '/'
                if source[-1] != sep:
                    self.deadTree.warning(
                        "Trailing separator missing! This might not be what you want!"
                    )

                # Make sure we get the top level directory, including folders, since
                # those can actually be deleted as well.
                self.bkpQueue.put((source, None, ['--filter=-,p */*']))

                # Crawl the top level (for now -- this is very basic, but decent performance)
                for item in listdir(source):
                    fullPath = join(source, item)

                    if isdir(fullPath) and not ismount(fullPath):
                        self.bkpQueue.put((source, item, None))

            # *default* is serial. Don't get confused!
            if not 'serial' in self.config['safestore'] or self.config['safestore']['serial']:
                self.bkpQueue.join()
                if hasattr(self, 'explosion'): 
                    raise self.explosion                
                else:
                    self.bkpComplete.set()
                    self.deadTree.info('Serial backup complete!')
            else:
                self.deadTree.info('Workers started for this backup!')

        if 'serial' in self.config['safestore'] and not self.config['safestore']['serial']:
            self.bkpQueue.join()
            if hasattr(self, 'explosion'):
                raise self.explosion
            else:
                self.bkpComplete.set()
                self.deadTree.info('Backup complete!')

    def rotate(self):
        self.deadTree.critical('Not implemented!')

    @asyncio.coroutine
    def _backup_worker(self, source, subDir=None, directArgs=None):
        '''The "meat" of the backup process

        Coroutine called by backup() for asynchronous use. This gets put in a task list, then run.
        '''
        pass